--#################################################################################################
-- Remove Events and jobs, drops of procs and functions are inline
--#################################################################################################
IF EXISTS ( SELECT * FROM sys.server_event_sessions WHERE name = 'EE_DDLChanges' )
  DROP EVENT SESSION [EE_DDLChanges] ON SERVER;
GO
USE [msdb];
GO
IF  EXISTS (SELECT * FROM [msdb].[dbo].[sysjobs] WHERE name = N'DBA_EE_DDLChanges')
  EXECUTE [msdb].[dbo].[sp_delete_job] @job_name=N'DBA_EE_DDLChanges', @delete_unused_schedule=1;
GO


--#################################################################################################
-- Create the extended event
--#################################################################################################
USE master;
IF NOT EXISTS(SELECT * FROM [sys].[server_event_sessions] AS [dxs] WHERE [dxs].[name] = 'EE_DDLChanges')
  BEGIN
    CREATE EVENT SESSION [EE_DDLChanges] ON SERVER 
ADD EVENT [sqlserver].[database_created](
    ACTION([sqlserver].[client_app_name],[sqlserver].[client_hostname],[sqlserver].[database_name],[sqlserver].[nt_username],[sqlserver].[sql_text],[sqlserver].[username])),
ADD EVENT [sqlserver].[object_altered](
    ACTION([sqlserver].[client_app_name],[sqlserver].[client_hostname],[sqlserver].[database_name],[sqlserver].[nt_username],[sqlserver].[sql_text],[sqlserver].[username])
    WHERE ([package0].[not_equal_uint64]([database_id],(2))       --NOT tempdb  avoids #temp table creation
    AND [package0].[equal_uint64]([ddl_phase],(1))                --Only Committed avoids duplicate statements/begin end
    AND [package0].[not_equal_uint64]([object_type],(17747))      --NOT SRVXESES avoids meaningless chatter
    --AND [package0].[not_equal_uint64]([object_type],(22601))      --NOT INDEX avoids meaningless chatter
    AND [package0].[not_equal_uint64]([object_type],(21587)))) ,  --NOT STATISTICS avoids meaningless chatter
ADD EVENT [sqlserver].[object_created](
    ACTION([sqlserver].[client_app_name],[sqlserver].[client_hostname],[sqlserver].[database_name],[sqlserver].[nt_username],[sqlserver].[sql_text],[sqlserver].[username])
    WHERE ([package0].[not_equal_uint64]([database_id],(2))       --NOT tempdb  avoids #temp table creation
    AND [package0].[equal_uint64]([ddl_phase],(1))                --Only Committed avoids duplicate statements/begin end
    AND [package0].[not_equal_uint64]([object_type],(17747))      --NOT SRVXESES avoids meaningless chatter
    --AND [package0].[not_equal_uint64]([object_type],(22601))      --NOT INDEX avoids meaningless chatter
    AND [package0].[not_equal_uint64]([object_type],(21587)))) ,  --NOT STATISTICS avoids meaningless chatter
ADD EVENT [sqlserver].[object_deleted](
 ACTION([sqlserver].[client_app_name],[sqlserver].[client_hostname],[sqlserver].[database_name],[sqlserver].[nt_username],[sqlserver].[sql_text],[sqlserver].[username])
     WHERE ([package0].[not_equal_uint64]([database_id],(2))      --NOT tempdb
    AND [package0].[equal_uint64]([ddl_phase],(1))                --Only Committed avoids duplicate statements/begin end
    AND [package0].[not_equal_uint64]([object_type],(17747))      --NOT SRVXESES avoids meaningless chatter
    --AND [package0].[not_equal_uint64]([object_type],(22601))      --NOT INDEX avoids meaningless chatter
    AND [package0].[not_equal_uint64]([object_type],(21587))))    --NOT STATISTICS avoids meaningless chatter
ADD TARGET [package0].[ring_buffer](SET max_memory=(1024000),occurrence_number=(1000))
WITH (MAX_MEMORY=4096 KB,
      EVENT_RETENTION_MODE=ALLOW_SINGLE_EVENT_LOSS,
      MAX_DISPATCH_LATENCY=30 SECONDS,
      MAX_EVENT_SIZE=0 KB,
      MEMORY_PARTITION_MODE=NONE,
      TRACK_CAUSALITY=OFF,
      STARTUP_STATE=ON)

--ADD TARGET package0.event_file(SET filename=N'EE_DDLChanges.xel',max_file_size=(200))
--            WITH (MAX_MEMORY=1024 KB,
--                  EVENT_RETENTION_MODE=ALLOW_SINGLE_EVENT_LOSS,
--                  MAX_DISPATCH_LATENCY=30 SECONDS,
--                  MAX_EVENT_SIZE=0 KB,
--                  MEMORY_PARTITION_MODE=NONE,
--                  TRACK_CAUSALITY=OFF,
--                  STARTUP_STATE=ON)
END
GO
IF NOT EXISTS (SELECT * FROM sys.[dm_xe_sessions] AS [dxs] WHERE [dxs].[name]='EE_DDLChanges')
ALTER EVENT SESSION [EE_DDLChanges] ON SERVER STATE = START
--ALTER EVENT SESSION [EE_DDLChanges] ON SERVER STATE = STOP
GO

USE [msdb];
GO
IF EXISTS(SELECT * FROM [sys].[objects] WHERE [object_id] = OBJECT_ID('[dbo].[fnExtendedEvent_DDLChanges]')) 
DROP FUNCTION [dbo].[fnExtendedEvent_DDLChanges]
GO
-- SELECT * FROM [dbo].[fnExtendedEvent_DDLChanges]() 
--#################################################################################################
-- EE_DDLChanges
--#################################################################################################
CREATE FUNCTION [dbo].[fnExtendedEvent_DDLChanges]() 
RETURNS @returntable TABLE 
(
[EventTimestamp]                  DATETIME2                                   NULL,
[EventName]                       VARCHAR(128)                                NULL,
[PackageName]                     VARCHAR(128)                                NULL,
[SessionIdentifier]               INT                                         NULL,
[PackageVersion]                  INT                                         NULL
-- data
,[CHANNEL]    VARCHAR(MAX)
,[collect_database_name]    BIT
,[database_id]    VARCHAR(MAX)
,[database_name]    VARCHAR(MAX)
,[ddl_phase]    VARCHAR(MAX)
,[index_id]    VARCHAR(MAX)
,[KEYWORD]    VARCHAR(MAX)
,[object_id]    INT
,[object_name]    VARCHAR(MAX)
,[object_type]    VARCHAR(MAX)
,[related_object_id]    INT
,[transaction_id]    VARCHAR(MAX)
,[UUID]    VARCHAR(MAX)
,[VERSION]    VARCHAR(MAX)

-- actions
,[action_client_app_name]    VARCHAR(MAX)
,[action_client_hostname]    VARCHAR(MAX)
,[action_database_name]    VARCHAR(MAX)
,[action_nt_username]    VARCHAR(MAX)
,[action_sql_text]    VARCHAR(MAX)
,[action_username]    VARCHAR(MAX)

)
AS 
BEGIN
DECLARE @xml XML;
SELECT  @xml = [tgtz].[target_data]
    FROM   [sys].[dm_xe_session_targets] [tgtz]
            JOIN [sys].[dm_xe_sessions] [sess]
              ON [tgtz].[event_session_address] = [sess].[address]
    WHERE  [sess].[name] = 'EE_DDLChanges'; 
INSERT INTO @returntable
SELECT 
[n].[value]('(@timestamp)[1]', 'datetime2') AS [EventTimestamp]
,[n].[value]('(@name)[1]', 'varchar(128)')     AS [EventName]
,[n].[value]('(@package)[1]', 'varchar(128)') AS [PackageName]
,[n].[value]('(@id)[1]', 'int')              AS [SessionIdentifier]
,[n].[value]('(@version)[1]', 'int')         AS [PackageVersion]
-- data
,[n].[value]('(data[@name="CHANNEL"]/value)[1]', 'varchar(max)') AS [data_CHANNEL]
,[n].[value]('(data[@name="collect_database_name"]/value)[1]', 'bit') AS [data_collect_database_name]
,[n].[value]('(data[@name="database_id"]/value)[1]', 'varchar(max)') AS [data_database_id]
,[n].[value]('(data[@name="database_name"]/value)[1]', 'varchar(max)') AS [data_database_name]
,[n].[value]('(data[@name="ddl_phase"]/value)[1]', 'varchar(max)') AS [data_ddl_phase]
,[n].[value]('(data[@name="index_id"]/value)[1]', 'varchar(max)') AS [data_index_id]
,[n].[value]('(data[@name="KEYWORD"]/value)[1]', 'varchar(max)') AS [data_KEYWORD]
,[n].[value]('(data[@name="object_id"]/value)[1]', 'int') AS [data_object_id]
,[n].[value]('(data[@name="object_name"]/value)[1]', 'varchar(max)') AS [data_object_name]
,[n].[value]('(data[@name="object_type"]/value)[1]', 'varchar(max)') AS [data_object_type]
,[n].[value]('(data[@name="related_object_id"]/value)[1]', 'int') AS [data_related_object_id]
,[n].[value]('(data[@name="transaction_id"]/value)[1]', 'varchar(max)') AS [data_transaction_id]
,[n].[value]('(data[@name="UUID"]/value)[1]', 'varchar(max)') AS [data_UUID]
,[n].[value]('(data[@name="VERSION"]/value)[1]', 'varchar(max)') AS [data_VERSION]

-- actions
,[n].[value]('(action[@name="client_app_name"]/value)[1]', 'varchar(max)') AS [action_client_app_name]
,[n].[value]('(action[@name="client_hostname"]/value)[1]', 'varchar(max)') AS [action_client_hostname]
,[n].[value]('(action[@name="database_name"]/value)[1]', 'varchar(max)') AS [action_database_name]
,[n].[value]('(action[@name="nt_username"]/value)[1]', 'varchar(max)') AS [action_nt_username]
,[n].[value]('(action[@name="sql_text"]/value)[1]', 'varchar(max)') AS [action_sql_text]
,[n].[value]('(action[@name="username"]/value)[1]', 'varchar(max)') AS [action_username]
FROM 
(SELECT
    @xml AS [event_data]
) AS [tab]
CROSS APPLY [event_data].nodes('RingBufferTarget/event') AS [q]([n])
----#################################################################################################
RETURN
END
GO

GO
 
USE [msdb];
GO
IF OBJECT_ID('[dbo].[DBA_EE_DDLChanges]') IS  NULL 
CREATE TABLE  [dbo].[DBA_EE_DDLChanges]
(
[EventTimestamp]                  DATETIME2                                   NULL,
[EventName]                       VARCHAR(128)                                NULL,
[PackageName]                     VARCHAR(128)                                NULL,
[SessionIdentifier]               INT                                         NULL,
[PackageVersion]                  INT                                         NULL,
[CHANNEL]    VARCHAR(MAX)
,[collect_database_name]    BIT
,[database_id]    VARCHAR(MAX)
,[database_name]    VARCHAR(MAX)
,[ddl_phase]    VARCHAR(MAX)
,[index_id]    VARCHAR(MAX)
,[KEYWORD]    VARCHAR(MAX)
,[object_id]    INT
,[object_name]    VARCHAR(MAX)
,[object_type]    VARCHAR(MAX)
,[related_object_id]    INT
,[transaction_id]    VARCHAR(MAX)
,[UUID]    VARCHAR(MAX)
,[VERSION]    VARCHAR(MAX)
,[action_client_app_name]    VARCHAR(MAX)
,[action_client_hostname]    VARCHAR(MAX)
,[action_database_name]    VARCHAR(MAX)
,[action_nt_username]    VARCHAR(MAX)
,[action_sql_text]    VARCHAR(MAX)
,[action_username]    VARCHAR(MAX)
)
GO
IF OBJECT_ID('[dbo].[usp_DBA_DDLChanges]') IS NOT NULL 
DROP  PROCEDURE [dbo].[usp_DBA_DDLChanges] 
GO

CREATE PROCEDURE [dbo].[usp_DBA_DDLChanges]
AS
BEGIN
--this procedure is doing blind inserts, but ideally you would want to change this
--to use insert update, based on specific columns to group by to avoid a billion row logging
INSERT INTO [dbo].[DBA_EE_DDLChanges]
SELECT *  FROM [dbo].[fnExtendedEvent_DDLChanges]() 
END
GO
USE [msdb];
GO
IF OBJECT_ID('[dbo].[usp_CYCLE_DDLChanges]') IS NOT NULL 
DROP  PROCEDURE [dbo].[usp_CYCLE_DDLChanges] 
GO

CREATE PROCEDURE [dbo].[usp_CYCLE_DDLChanges]  @SetStatus VARCHAR(30) 
AS
BEGIN --PROC
DECLARE @ee VARCHAR(128) = 'EE_DDLChanges';
 BEGIN
  IF @SetStatus = 'OFF'
    BEGIN
     --try it twice
      IF EXISTS (SELECT COUNT(name) FROM sys.dm_xe_sessions WHERE name = @ee GROUP BY name HAVING COUNT(name) = 1 )
        ALTER EVENT SESSION [EE_DDLChanges] ON SERVER STATE = STOP;
        WAITFOR DELAY '00:00:01'
       IF EXISTS (SELECT COUNT(name) FROM sys.dm_xe_sessions WHERE name = @ee GROUP BY name HAVING COUNT(name) = 1 )
        ALTER EVENT SESSION [EE_DDLChanges] ON SERVER STATE = STOP;
    END
   ELSE --@SetStatus = ON -turn it on
     BEGIN
       --try it twice
      IF NOT EXISTS (SELECT COUNT(name) FROM sys.dm_xe_sessions WHERE name = @ee GROUP BY name HAVING COUNT(name) = 1 )
        ALTER EVENT SESSION [EE_DDLChanges] ON SERVER STATE = START;
        WAITFOR DELAY '00:00:01'
       IF NOT EXISTS (SELECT COUNT(name) FROM sys.dm_xe_sessions WHERE name = @ee GROUP BY name HAVING COUNT(name) = 1 )
        ALTER EVENT SESSION [EE_DDLChanges] ON SERVER STATE = START;
     END
 END;
END; -- PROC
GO

GO

--#################################################################################################
-- Create the Job
--#################################################################################################
USE [msdb]
GO
DECLARE @saAccount varchar(128)
SELECT @saAccount = name from sys.server_principals WHERE sid = 0x01
PRINT 'sa account is ' + @saAccount
/****** Object:  Job [DBA_EE_DDLChanges] ******/
/****** Object:  JobCategory [Data Collector] ******/
IF NOT EXISTS (SELECT [name] FROM [msdb].[dbo].[syscategories] WHERE [name]=N'Data Collector' AND [category_class]=1)
BEGIN
EXEC [msdb].[dbo].[sp_add_category] @class=N'JOB', @type=N'LOCAL', @name=N'Data Collector'
END
IF NOT EXISTS (SELECT name FROM msdb.dbo.sysjobs WHERE name=N'DBA_EE_DDLChanges' )
BEGIN
DECLARE @jobId BINARY(16)
EXEC [msdb].[dbo].[sp_add_job] @job_name=N'DBA_EE_DDLChanges', 
@enabled=1, 
@notify_level_eventlog=3, 
@notify_level_email=2, 
@notify_level_netsend=0, 
@notify_level_page=0, 
@delete_level=0, 
@description=N'Logs Events into DBA_EE_DDLChanges', 
@category_name=N'Data Collector', 
@owner_login_name=@saAccount --
--,@notify_email_operator_name=N'admin@stormrage.com'

/****** Object:  Step [LogEventErrorCapture] ******/
EXEC [msdb].[dbo].[sp_add_jobstep] @job_name=N'DBA_EE_DDLChanges', @step_name=N'DBA_EE_DDLChanges', 
@step_id=1, 
@cmdexec_success_code=0, 
@on_success_action=3, 
@on_success_step_id=0, 
@on_fail_action=2, 
@on_fail_step_id=0, 
@retry_attempts=0, 
@retry_interval=0, 
@os_run_priority=0, @subsystem=N'TSQL', 
@command=N'
EXEC [dbo].[usp_DBA_DDLChanges]
', 
    @database_name=N'msdb', 
    @flags=4

/****** Object:  Step [CycleDDLChanges OFF] ******/
EXEC [msdb].[dbo].[sp_add_jobstep] @job_name=N'DBA_EE_DDLChanges', @step_name=N'CycleDDLChanges OFF', 
@step_id=2, 
@cmdexec_success_code=0, 
@on_success_action=3, 
@on_success_step_id=0, 
@on_fail_action=2, 
@on_fail_step_id=0, 
@retry_attempts=0, 
@retry_interval=0, 
@os_run_priority=0, @subsystem=N'TSQL', 
@command=N'EXEC usp_Cycle_DDLChanges  @SetStatus = ''OFF''', 
@database_name=N'msdb', 
@flags=4

/****** Object:  Step [CycleDDLChanges ON]  ******/
EXEC [msdb].[dbo].[sp_add_jobstep] @job_name=N'DBA_EE_DDLChanges', @step_name=N'CycleDDLChanges ON', 
@step_id=3, 
@cmdexec_success_code=0, 
@on_success_action=1, 
@on_success_step_id=0, 
@on_fail_action=2, 
@on_fail_step_id=0, 
@retry_attempts=0, 
@retry_interval=0, 
@os_run_priority=0,
@subsystem=N'TSQL', 
@command=N'Exec usp_Cycle_DDLChanges  @SetStatus = ''ON''', 
@database_name=N'msdb', 
@flags=0

EXEC [msdb].[dbo].[sp_update_job] @job_name=N'DBA_EE_DDLChanges', @start_step_id = 1

EXEC  [msdb].[dbo].[sp_add_jobschedule] @job_name=N'DBA_EE_DDLChanges', @name=N'DBA_EE_DDLChanges', 
@enabled=1, 
@freq_type=4, 
@freq_interval=1, 
@freq_subday_type=4, 
@freq_subday_interval=5, 
@freq_relative_interval=0, 
@freq_recurrence_factor=0, 
@active_start_date=20140110, 
@active_end_date=99991231, 
@active_start_time=0, 
@active_end_time=235959, 
@schedule_uid=N'a53a7018-0bc1-4975-976f-e0d55451c5b3'

EXEC [msdb].[dbo].[sp_add_jobserver] @job_name=N'DBA_EE_DDLChanges', @server_name = N'(local)'
END 

USE [msdb]


/****** Object:  Job [[DBA]_DDLChangeLog_ManageHistory]    Script Date: 9/13/2021 10:47:06 AM ******/
/****** Object:  JobCategory [[Uncategorized (Local)]]    Script Date: 9/13/2021 10:47:06 AM ******/
IF NOT EXISTS (SELECT name FROM msdb.dbo.syscategories WHERE name=N'[Uncategorized (Local)]' AND category_class=1)
BEGIN
EXEC msdb.dbo.sp_add_category @class=N'JOB', @type=N'LOCAL', @name=N'[Uncategorized (Local)]'
END

IF NOT EXISTS (SELECT name FROM msdb.dbo.sysjobs WHERE name=N'[DBA]_DDLChangeLog_ManageHistory' )
BEGIN
EXEC msdb.dbo.sp_add_job @job_name=N'[DBA]_DDLChangeLog_ManageHistory', 
		@enabled=1, 
		@notify_level_eventlog=0, 
		@notify_level_email=0, 
		@notify_level_netsend=0, 
		@notify_level_page=0, 
		@delete_level=0, 
		@description=N'No description available.', 
		@category_name=N'[Uncategorized (Local)]', 
		@owner_login_name=@saAccount

/****** Object:  Step [[DBA]_DDLChangeLog_ManageHistory]    Script Date: 9/13/2021 10:47:07 AM ******/
EXEC msdb.dbo.sp_add_jobstep @job_name=N'[DBA]_DDLChangeLog_ManageHistory', @step_name=N'[DBA]_DDLChangeLog_ManageHistory', 
		@step_id=1, 
		@cmdexec_success_code=0, 
		@on_success_action=1, 
		@on_success_step_id=0, 
		@on_fail_action=2, 
		@on_fail_step_id=0, 
		@retry_attempts=0, 
		@retry_interval=0, 
		@os_run_priority=0, @subsystem=N'TSQL', 
		@command=N'--one year plus 95 days old, delete it.
    IF EXISTS(SELECT * FROM sys.tables WHERE object_id = OBJECT_ID(''[dbo].[DDLChangeLog]''))
      EXECUTE(''DELETE FROM [dbo].[DDLChangeLog] WHERE CreateDate < DATEADD(dd,-460,GETDATE())'')
    IF EXISTS(SELECT * FROM sys.tables WHERE object_id = OBJECT_ID(''[dbo].[DBA_EE_DDLChanges]''))
      EXECUTE(''DELETE FROM [dbo].[DBA_EE_DDLChanges] WHERE EventTimestamp < DATEADD(dd,-460,GETDATE())'')', 
		@database_name=N'msdb', 
		@flags=0

EXEC msdb.dbo.sp_update_job @job_name=N'[DBA]_DDLChangeLog_ManageHistory', @start_step_id = 1

EXEC msdb.dbo.sp_add_jobschedule @job_name=N'[DBA]_DDLChangeLog_ManageHistory', @name=N'Once a Week', 
		@enabled=1, 
		@freq_type=8, 
		@freq_interval=1, 
		@freq_subday_type=1, 
		@freq_subday_interval=0, 
		@freq_relative_interval=0, 
		@freq_recurrence_factor=1, 
		@active_start_date=20210903, 
		@active_end_date=99991231, 
		@active_start_time=40000, 
		@active_end_time=235959, 
		@schedule_uid=N'd243455d-1d57-41bd-9061-4c618a7e8912'

EXEC msdb.dbo.sp_add_jobserver @job_name=N'[DBA]_DDLChangeLog_ManageHistory', @server_name = N'(local)'
END

