USE master;
GO
IF  EXISTS (SELECT * FROM [msdb].[dbo].[sysjobs] WHERE name = N'DBA_LoginTracking')
  EXECUTE [msdb].[dbo].[sp_delete_job] @job_name=N'DBA_LoginTracking', @delete_unused_schedule=1;

--#################################################################################################
-- Create The Extended Event
--#################################################################################################

IF  EXISTS(SELECT * FROM sys.[dm_xe_sessions] WHERE name = 'LoginTracking')
OR  EXISTS(SELECT * FROM sys.[server_event_sessions]  WHERE name = 'LoginTracking')
DROP EVENT SESSION [LoginTracking] ON SERVER 
GO
IF NOT EXISTS(SELECT * FROM sys.[dm_xe_sessions] WHERE name = 'LoginTracking')
AND NOT EXISTS(SELECT * FROM sys.[server_event_sessions]  WHERE name = 'LoginTracking')

CREATE EVENT SESSION [LoginTracking] ON SERVER 
ADD EVENT sqlserver.connectivity_ring_buffer_recorded(
    ACTION(sqlserver.client_app_name,sqlserver.client_hostname,sqlserver.server_principal_name,sqlserver.session_id,sqlserver.username)),
ADD EVENT sqlserver.login(SET collect_options_text=(1)
    ACTION(sqlserver.client_app_name,sqlserver.client_hostname,sqlserver.server_instance_name,sqlserver.server_principal_name,sqlserver.username))
ADD TARGET package0.ring_buffer
WITH (MAX_MEMORY=4096 KB,EVENT_RETENTION_MODE=ALLOW_SINGLE_EVENT_LOSS,MAX_DISPATCH_LATENCY=30 SECONDS,MAX_EVENT_SIZE=0 KB,MEMORY_PARTITION_MODE=NONE,TRACK_CAUSALITY=ON,STARTUP_STATE=ON)
GO
ALTER EVENT SESSION [LoginTracking] ON SERVER STATE=START;
GO
USE msdb
--#################################################################################################
-- Create The matching Funciton
--#################################################################################################
IF OBJECT_ID('[dbo].[fnExtendedEvent_LoginTracking]') IS NOT NULL 
DROP  FUNCTION  [dbo].[fnExtendedEvent_LoginTracking] 
GO
--SELECT * into #LoginTracking FROM [dbo].[fnExtendedEvent_LoginTracking]() 
--#################################################################################################
-- LoginTracking
--#################################################################################################
CREATE FUNCTION [dbo].[fnExtendedEvent_LoginTracking]() 
RETURNS @returntable TABLE 
(
[EventTimestamp]                  DATETIME2                                   NULL,
[EventName]                       VARCHAR(128)                                NULL,
[PackageName]                     VARCHAR(128)                                NULL,
[SessionIdentifier]               INT                                         NULL,
[PackageVersion]                  INT                                         NULL
-- data
,[collect_database_name]    bit
,[collect_options_text]    bit
,[connection_id]    varchar(max)
,[exec_classifier_ms]    varchar(max)
,[is_cached]    bit
,[is_client]    bit
,[is_dac]    bit
,[is_recovered]    bit
,[local_host]    varchar(max)
,[login_task_enqueued_ms]    varchar(max)
,[logon_triggers_ms]    varchar(max)
,[options]    varchar(max)
,[options_text]    varchar(max)
,[packet_size]    varchar(max)
,[remote_host]    varchar(max)
,[session_recover_ms]    varchar(max)
,[ssl_enqueue_ms]    varchar(max)
,[ssl_net_writes_ms]    varchar(max)
,[ssl_secure_calls_ms]    varchar(max)
,[sspi_net_reads_ms]    varchar(max)
,[sspi_processing_ms]    varchar(max)
,[state]    varchar(max)
,[tds_input_buffer_bytes]    varchar(max)
,[tds_output_buffer_error]    varchar(max)

-- actions
,[client_hostname]    varchar(max)
,[server_instance_name]    varchar(max)
,[server_principal_name]    varchar(max)
,[session_id]    varchar(max)

)
AS 
BEGIN
DECLARE @xml XML;
SELECT  @xml = tgtz.target_data
    FROM   sys.dm_xe_session_targets tgtz
            JOIN sys.dm_xe_sessions sess
              ON tgtz.event_session_address = sess.address
    WHERE  sess.name = 'LoginTracking'; 
INSERT INTO @returntable
SELECT 
n.value('(@timestamp)[1]', 'datetime2') AS [EventTimestamp]
,n.value('(@name)[1]', 'varchar(128)')     AS EventName
,n.value('(@package)[1]', 'varchar(128)') AS PackageName
,n.value('(@id)[1]', 'int')              AS SessionIdentifier
,n.value('(@version)[1]', 'int')         AS PackageVersion
-- data
,n.value('(data[@name="collect_database_name"]/value)[1]', 'bit') AS data_collect_database_name
,n.value('(data[@name="collect_options_text"]/value)[1]', 'bit') AS data_collect_options_text
,n.value('(data[@name="connection_id"]/value)[1]', 'varchar(max)') AS data_connection_id
,n.value('(data[@name="exec_classifier_ms"]/value)[1]', 'varchar(max)') AS data_exec_classifier_ms
,n.value('(data[@name="is_cached"]/value)[1]', 'bit') AS data_is_cached
,n.value('(data[@name="is_client"]/value)[1]', 'bit') AS data_is_client
,n.value('(data[@name="is_dac"]/value)[1]', 'bit') AS data_is_dac
,n.value('(data[@name="is_recovered"]/value)[1]', 'bit') AS data_is_recovered
,n.value('(data[@name="local_host"]/value)[1]', 'varchar(max)') AS data_local_host
,n.value('(data[@name="login_task_enqueued_ms"]/value)[1]', 'varchar(max)') AS data_login_task_enqueued_ms
,n.value('(data[@name="logon_triggers_ms"]/value)[1]', 'varchar(max)') AS data_logon_triggers_ms
,n.value('(data[@name="options"]/value)[1]', 'varchar(max)') AS data_options
,n.value('(data[@name="options_text"]/value)[1]', 'varchar(max)') AS data_options_text
,n.value('(data[@name="packet_size"]/value)[1]', 'varchar(max)') AS data_packet_size
,n.value('(data[@name="remote_host"]/value)[1]', 'varchar(max)') AS data_remote_host
,n.value('(data[@name="session_recover_ms"]/value)[1]', 'varchar(max)') AS data_session_recover_ms
,n.value('(data[@name="ssl_enqueue_ms"]/value)[1]', 'varchar(max)') AS data_ssl_enqueue_ms
,n.value('(data[@name="ssl_net_writes_ms"]/value)[1]', 'varchar(max)') AS data_ssl_net_writes_ms
,n.value('(data[@name="ssl_secure_calls_ms"]/value)[1]', 'varchar(max)') AS data_ssl_secure_calls_ms
,n.value('(data[@name="sspi_net_reads_ms"]/value)[1]', 'varchar(max)') AS data_sspi_net_reads_ms
,n.value('(data[@name="sspi_processing_ms"]/value)[1]', 'varchar(max)') AS data_sspi_processing_ms
,n.value('(data[@name="state"]/value)[1]', 'varchar(max)') AS data_state
,n.value('(data[@name="tds_input_buffer_bytes"]/value)[1]', 'varchar(max)') AS data_tds_input_buffer_bytes
,n.value('(data[@name="tds_output_buffer_error"]/value)[1]', 'varchar(max)') AS data_tds_output_buffer_error

-- actions
,n.value('(action[@name="client_hostname"]/value)[1]', 'varchar(max)') AS action_client_hostname
,n.value('(action[@name="server_instance_name"]/value)[1]', 'varchar(max)') AS action_server_instance_name
,n.value('(action[@name="server_principal_name"]/value)[1]', 'varchar(max)') AS action_server_principal_name
,n.value('(action[@name="session_id"]/value)[1]', 'varchar(max)') AS action_session_id
FROM 
(SELECT
    @xml AS event_data
) as tab
CROSS APPLY event_data.nodes('RingBufferTarget/event') as q(n)
----#################################################################################################
RETURN
END
GO
IF OBJECT_ID('[dbo].[usp_DBALoginTracking]') IS NOT NULL 
DROP  PROCEDURE [dbo].[usp_DBALoginTracking] 
GO
CREATE PROCEDURE [dbo].[usp_DBALoginTracking]
AS
BEGIN
IF EXISTS(SELECT name FROM sys.columns WHERE OBJECT_NAME(object_id)  = 'DBALoginTracking' AND column_id = 1 AND name <> 'FirstTrackedLoginDate')
EXECUTE ('DROP TABLE [dbo].[DBALoginTracking] ;')
IF OBJECT_ID('[dbo].[DBALoginTracking]') IS  NULL 
EXECUTE('
CREATE TABLE [dbo].[DBALoginTracking] ( 
[FirstTrackedLoginDate]  DATETIME2                                                NULL,
[LastTrackedLoginDate]   DATETIME2                                                NULL,
[EventName]              VARCHAR(128)                                             NULL,
[PackageName]            VARCHAR(128)                                             NULL,
[options]                VARCHAR(max)                                             NULL,
[options_text]           VARCHAR(max)                                             NULL,
[packet_size]            VARCHAR(max)                                             NULL,
[remote_host]            VARCHAR(max)                                             NULL,
[client_hostname]        VARCHAR(max)                                             NULL,
[server_instance_name]   VARCHAR(max)                                             NULL,
[server_principal_name]  VARCHAR(max)                                             NULL)
')
IF OBJECT_ID('tempdb..[#LoginTracking]') IS NOT NULL 
DROP TABLE [#LoginTracking] 
SELECT * into #LoginTracking FROM [dbo].[fnExtendedEvent_LoginTracking]() 
--update existing rows with the "Last Seen"

UPDATE MyTarget
SET MyTarget.LastTrackedLoginDate = MySource.[EventTimestamp]
FROM  msdb.dbo.DBALoginTracking MyTarget 
INNER JOIN (SELECT 
              MAX([EventTimestamp]) AS [EventTimestamp],
              [client_hostname],
              [server_principal_name] 
            FROM #LoginTracking 
            GROUP BY 
              [client_hostname],
              [server_principal_name]) MySource
ON MySource.[client_hostname] = MyTarget.[client_hostname]
AND MySource.[server_principal_name] = MyTarget.[server_principal_name]
WHERE MySource.[client_hostname] = MyTarget.[client_hostname]
AND MySource.[server_principal_name] = MyTarget.[server_principal_name]
AND MyTarget.LastTrackedLoginDate < MySource.[EventTimestamp]

--insert any new machine/login combos
INSERT INTO msdb.dbo.DBALoginTracking([FirstTrackedLoginDate],[LastTrackedLoginDate],[EventName],[PackageName],[options],[options_text],[packet_size],[remote_host],[client_hostname],[server_instance_name],[server_principal_name])
SELECT 
MIN(MySource.[EventTimestamp])      AS FirstTrackedLoginDate,
MAX(MySource.[EventTimestamp])      AS LastTrackedLoginDate,
MAX(MySource.[EventName])           AS [EventName],
MAX(MySource.[PackageName])         AS [PackageName],
MAX(MySource.[options])             AS [options],
MAX(MySource.[options_text])        AS [options_text],
MAX(MySource.[packet_size])         AS [packet_size],
MAX(MySource.[remote_host])         AS [remote_host],
MySource.[client_hostname],
MAX(MySource.[server_instance_name])AS [server_instance_name],
MySource.[server_principal_name]
FROM  #LoginTracking MySource
LEFT JOIN msdb.dbo.DBALoginTracking MyTarget
ON MySource.[client_hostname] = MyTarget.[client_hostname]
AND MySource.[server_principal_name] = MyTarget.[server_principal_name]
WHERE MyTarget.[server_principal_name] IS NULL
GROUP BY 
MySource.[client_hostname],
MySource.[server_principal_name]
--toggle the ring buffer.
END
GO
--#################################################################################################
-- 
--#################################################################################################
IF OBJECT_ID('[dbo].[usp_CYCLE_LoginTracking]') IS NOT NULL 
DROP  PROCEDURE [dbo].[usp_CYCLE_LoginTracking] 
GO
--#################################################################################################
--  PROCEDURE [dbo].[usp_CYCLE_LoginTracking]
--#################################################################################################
CREATE PROCEDURE [dbo].[usp_CYCLE_LoginTracking]  @SetStatus VARCHAR(30) 
AS
BEGIN --PROC
DECLARE @ee VARCHAR(128) = 'LoginTracking';
 BEGIN
  IF @SetStatus = 'OFF'
    BEGIN
     --try it twice
      IF EXISTS (SELECT COUNT(name) FROM sys.dm_xe_sessions WHERE name = @ee GROUP BY name HAVING COUNT(name) = 1 )
        ALTER EVENT SESSION [LoginTracking] ON SERVER STATE = STOP;
        WAITFOR DELAY '00:00:01'
       IF EXISTS (SELECT COUNT(name) FROM sys.dm_xe_sessions WHERE name = @ee GROUP BY name HAVING COUNT(name) = 1 )
        ALTER EVENT SESSION [LoginTracking] ON SERVER STATE = STOP;
    END
   ELSE --@SetStatus = ON -turn it on
     BEGIN
       --try it twice
      IF NOT EXISTS (SELECT COUNT(name) FROM sys.dm_xe_sessions WHERE name = @ee GROUP BY name HAVING COUNT(name) = 1 )
        ALTER EVENT SESSION [LoginTracking] ON SERVER STATE = START;
        WAITFOR DELAY '00:00:01'
       IF NOT EXISTS (SELECT COUNT(name) FROM sys.dm_xe_sessions WHERE name = @ee GROUP BY name HAVING COUNT(name) = 1 )
        ALTER EVENT SESSION [LoginTracking] ON SERVER STATE = START;
     END
 END;
END; -- PROC
GO
--#################################################################################################
-- Create the Operator
--#################################################################################################
IF NOT EXISTS(SELECT * FROM [msdb].[dbo].[sysoperators] WHERE name = 'admin@stormrage.com')
BEGIN
  EXECUTE ('EXEC msdb.dbo.sp_add_operator 
    @name=N''admin@stormrage.com'', 
    @enabled=1, 
    @weekday_pager_start_time=90000, 
    @weekday_pager_end_time=180000, 
    @saturday_pager_start_time=90000, 
    @saturday_pager_end_time=180000, 
    @sunday_pager_start_time=90000, 
    @sunday_pager_end_time=180000, 
    @pager_days=0, 
    @email_address=N''admin@stormrage.com'', 
    @category_name=N''[Uncategorized]''');
END
--#################################################################################################
-- Create the Job
--#################################################################################################
USE [msdb]
GO
IF EXISTS(SELECT * FROM msdb.dbo.sysjobs WHERE name = N'DBA_LoginTracking')
EXECUTE msdb.dbo.[sp_delete_job] @job_name = N'DBA_LoginTracking', @delete_unused_schedule = 1
GO
DECLARE @saAccount varchar(128)
SELECT @saAccount = name from sys.server_principals WHERE sid = 0x01
PRINT 'sa account is ' + @saAccount
/****** Object:  Job [DBA_LoginTracking]    Script Date: 9/16/2020 3:12:42 PM ******/
BEGIN TRANSACTION
DECLARE @ReturnCode INT
SELECT @ReturnCode = 0
/****** Object:  JobCategory [Data Collector]    Script Date: 9/16/2020 3:12:42 PM ******/
IF NOT EXISTS (SELECT name FROM msdb.dbo.syscategories WHERE name=N'Data Collector' AND category_class=1)
BEGIN
EXEC @ReturnCode = msdb.dbo.sp_add_category @class=N'JOB', @type=N'LOCAL', @name=N'Data Collector'
IF (@@ERROR <> 0 OR @ReturnCode <> 0) GOTO QuitWithRollback

END

DECLARE @jobId BINARY(16)
EXEC @ReturnCode =  msdb.dbo.sp_add_job @job_name=N'DBA_LoginTracking', 
    @enabled=1, 
    @notify_level_eventlog=3, 
    @notify_level_email=2, 
    @notify_level_netsend=0, 
    @notify_level_page=0, 
    @delete_level=0, 
    @description=N'Logs Events into DBALoginTracking', 
    @category_name=N'Data Collector', 
    @owner_login_name=@saAccount, 
    @notify_email_operator_name=N'admin@stormrage.com', @job_id = @jobId OUTPUT
IF (@@ERROR <> 0 OR @ReturnCode <> 0) GOTO QuitWithRollback
/****** Object:  Step [LogEventErrorCapture]    Script Date: 9/16/2020 3:12:43 PM ******/
EXEC @ReturnCode = msdb.dbo.sp_add_jobstep @job_id=@jobId, @step_name=N'TrackLogins', 
    @step_id=1, 
    @cmdexec_success_code=0, 
    @on_success_action=3, 
    @on_success_step_id=0, 
    @on_fail_action=2, 
    @on_fail_step_id=0, 
    @retry_attempts=0, 
    @retry_interval=0, 
    @os_run_priority=0, @subsystem=N'TSQL', 
    @command=N'
        EXEC [dbo].[usp_DBALoginTracking]
', 
    @database_name=N'msdb', 
    @flags=4
IF (@@ERROR <> 0 OR @ReturnCode <> 0) GOTO QuitWithRollback
/****** Object:  Step [CycleLoginTracking OFF]    Script Date: 9/16/2020 3:12:43 PM ******/
EXEC @ReturnCode = msdb.dbo.sp_add_jobstep @job_id=@jobId, @step_name=N'CycleLoginTracking OFF', 
    @step_id=2, 
    @cmdexec_success_code=0, 
    @on_success_action=3, 
    @on_success_step_id=0, 
    @on_fail_action=2, 
    @on_fail_step_id=0, 
    @retry_attempts=0, 
    @retry_interval=0, 
    @os_run_priority=0, @subsystem=N'TSQL', 
    @command=N'EXEC usp_Cycle_LoginTracking  @SetStatus = ''OFF''', 
    @database_name=N'msdb', 
    @flags=4
IF (@@ERROR <> 0 OR @ReturnCode <> 0) GOTO QuitWithRollback
/****** Object:  Step [CycleLoginTracking ON]    Script Date: 9/16/2020 3:12:43 PM ******/
EXEC @ReturnCode = msdb.dbo.sp_add_jobstep @job_id=@jobId, @step_name=N'CycleLoginTracking ON', 
    @step_id=3, 
    @cmdexec_success_code=0, 
    @on_success_action=1, 
    @on_success_step_id=0, 
    @on_fail_action=2, 
    @on_fail_step_id=0, 
    @retry_attempts=0, 
    @retry_interval=0, 
    @os_run_priority=0, @subsystem=N'TSQL', 
    @command=N'Exec usp_Cycle_LoginTracking  @SetStatus = ''ON''', 
    @database_name=N'msdb', 
    @flags=0
IF (@@ERROR <> 0 OR @ReturnCode <> 0) GOTO QuitWithRollback
EXEC @ReturnCode = msdb.dbo.sp_update_job @job_id = @jobId, @start_step_id = 1
IF (@@ERROR <> 0 OR @ReturnCode <> 0) GOTO QuitWithRollback
EXEC @ReturnCode = msdb.dbo.sp_add_jobschedule @job_id=@jobId, @name=N'DBALoginTracking', 
    @enabled=1, 
    @freq_type=4, 
    @freq_interval=1, 
    @freq_subday_type=4, 
    @freq_subday_interval=5, 
    @freq_relative_interval=0, 
    @freq_recurrence_factor=0, 
    @active_start_date=20140110, 
    @active_end_date=99991231, 
    @active_start_time=0, 
    @active_end_time=235959, 
    @schedule_uid=N'a53a7018-0bc1-4975-976f-e0d55451c5b3'
IF (@@ERROR <> 0 OR @ReturnCode <> 0) GOTO QuitWithRollback
EXEC @ReturnCode = msdb.dbo.sp_add_jobserver @job_id = @jobId, @server_name = N'(local)'
IF (@@ERROR <> 0 OR @ReturnCode <> 0) GOTO QuitWithRollback
COMMIT TRANSACTION
GOTO EndSave
QuitWithRollback:
    IF (@@TRANCOUNT > 0) ROLLBACK TRANSACTION
EndSave:
GO


