--#################################################################################################
-- 2025-06-03 06:27:37.513 MachineName:ViscosityLowell; ComputerBiosName:VISCOSITYLOWELL;  
-- ApplicationName: DBA | AppEmailGroup: lowell.izaguirre@viscosityna.com | ServerUptime[dd:hh:mm:ss]: 01:23:43:24
-- Cname: wks-lowell.workgroup,;  LoginName: VISCOSITYLOWELL\lowell Authorization Scheme: NTLM viscositylowell
-- Context: ViscosityLowell,; Database:master; Version: [SQL 2022 || 16.0.4165.4] Developer Edition (64-bit)
--#################################################################################################
SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED;
GO
--desc: Reviews all compiled objects to try and coutn the number of Common Table Expressions in it's definition
--exessive nested CTE's are an opportunity for Divide And conquer style to pull out data to temp tables to increase performance
IF OBJECT_ID('tempdb..#ModuleDefinitions') IS NOT NULL DROP TABLE #ModuleDefinitions;
CREATE TABLE #ModuleDefinitions(
[Observation]   VARCHAR(30)                             NOT NULL,
[DatabaseName]  SYSNAME                                 NOT NULL,
[SchemaName]    SYSNAME                                 NOT NULL,
[ObjectName]     SYSNAME                                 NOT NULL,
[ObjectType]    NVARCHAR(128)                               NULL,
[Definition]    NVARCHAR(max)                               NULL)

EXECUTE sp_MSforeachdb '
INSERT INTO #ModuleDefinitions([Observation],[DatabaseName],[SchemaName],[ObjectName],[ObjectType],[Definition])
SELECT ''Nested CTE Review'' AS Observation,
  ''?'' AS DatabaseName,
  schmz.name As SchemaName,
  tabz.name AS [ObjectName],
  tabz.type_desc AS ObjectType,
  def.[definition]
FROM [?].sys.objects tabz
  INNER JOIN [?].sys.schemas schmz ON tabz.schema_id = schmz.schema_id
  INNER JOIN [?].sys.sql_modules def  ON tabz.object_id = def.object_id
'

-- Update #ModuleDefinitions to strip multi-line comments
UPDATE #ModuleDefinitions
SET Definition = CASE 
    WHEN CHARINDEX('/*', Definition) > 0 AND CHARINDEX('*/', Definition) > CHARINDEX('/*', Definition)
    THEN STUFF(
        Definition,
        CHARINDEX('/*', Definition),
        CHARINDEX('*/', Definition) - CHARINDEX('/*', Definition) + 2,
        ''
    )
    ELSE Definition
END;

-- Update #ModuleDefinitions to strip single-line comments
UPDATE #ModuleDefinitions
SET Definition = CASE 
    WHEN CHARINDEX('--', Definition) > 0 
    THEN STUFF(
        Definition,
        CHARINDEX('--', Definition),
        CHARINDEX(CHAR(10), Definition + CHAR(10), CHARINDEX('--', Definition)) - CHARINDEX('--', Definition),
        ''
    )
    ELSE Definition
END;
UPDATE #ModuleDefinitions SET definition = REPLACE(definition, CHAR(13) + CHAR(10), ' '); -- Remove line breaks
UPDATE #ModuleDefinitions SET definition = REPLACE(definition, CHAR(9), ' '); -- Remove tabs
UPDATE #ModuleDefinitions SET definition = REPLACE(definition, CHAR(13) , ' '); -- Remove line breaks
UPDATE #ModuleDefinitions SET definition = REPLACE(definition, CHAR(10), ' '); -- Remove line breaks
UPDATE #ModuleDefinitions SET definition = REPLACE(definition, ' ', ''); -- Remove all spaces
-- Query to find modules with nested CTEs using CROSS APPLY
SELECT 

    (LEN(md.Definition) - LEN(REPLACE(md.Definition, 'AS(', ''))) / LEN('AS(') AS NestedCTECount,
        md.*
FROM 
    #ModuleDefinitions AS md
WHERE 
    (LEN(md.Definition) - LEN(REPLACE(md.Definition, 'AS(', ''))) / LEN('AS(') > 0
ORDER BY NestedCTECount desc


-- Clean up temporary table
DROP TABLE #ModuleDefinitions;