--desc: best practices for static code scan
IF OBJECT_ID('[dbo].[StripComments]') IS NOT NULL 
DROP  FUNCTION [dbo].[StripComments]; 
GO
--#################################################################################################
-- Real World DBA Toolkit Version 2019-06-01 Lowell Izaguirre lowell@stormrage.com
--#################################################################################################
CREATE FUNCTION [dbo].[StripComments](@StringToStrip VARCHAR(MAX))
RETURNS VARCHAR(MAX)
WITH SCHEMABINDING
AS
BEGIN
    -- This bit strips out block comments.  We need to strip them out BEFORE
    -- single line comments (like this one), because you could theoretically have
    -- a block comment like this:
    /* My Comment
    -- Is malformed */
    -- Variables to hold the first and last character's positions in the "next" block 
    -- comment in the string
    SET @StringToStrip = REPLACE(@StringToStrip,CHAR(13),'');
    DECLARE @CodeBlockStart INT, @CodeBlockEnd INT;
    SET @CodeBlockStart = PATINDEX(CHAR(37) + CHAR(47) + CHAR(42)  + CHAR(37), @StringToStrip); --this is %/*%
    -- Loop as long as we still have comments to exorcise ;)
    WHILE @CodeBlockStart > 0
    BEGIN
        -- Grab the last character in the code block by searching for the first incidence
        -- of */ (close comment) in the string.
        SET @CodeBlockEnd = PATINDEX(CHAR(37) + CHAR(42)  + CHAR(47) +  CHAR(37) , @StringToStrip); --this is %*/%
       --is there a ghost slash asterisk without a match?
        IF @CodeBlockEnd = 0 
        BEGIN -- if we get here the SQL dosn't have a closing '*/' -- So we just delete it 
        SET @CodeBlockEnd = @CodeBlockStart + 2; 
        END;
        ELSE IF @CodeBlockEnd < @CodeBlockStart -- if we get a */ before /* we delete it
        BEGIN
        SET @CodeBlockStart = @CodeBlockEnd -2;
        END;
        -- "Cut" out the comment by concatenating everything the the "left" and "right"
        -- of the comment
        SET @StringToStrip = LEFT(@StringToStrip, @CodeBlockStart - 1)
                                + RIGHT(@StringToStrip, LEN(@StringToStrip) - (@CodeBlockEnd + 1));
    
        -- Fetch the first character's position in the next comment block, if there is one.
        SET @CodeBlockStart = PATINDEX(CHAR(37) + CHAR(47) + CHAR(42)  + CHAR(37), @StringToStrip);
    END;
    -- Once code blocks are out, we can remove any lines commented by double dashes (like this one)
    -- Variables to hold the first and last character's position in the "next" code block.
    DECLARE @DoubleDashStart INT, @DoubleDashLineEnd INT;
    -- Grab the first double-dash (if there is one)
    SET @DoubleDashStart = PATINDEX(CHAR(37) + CHAR(45) + CHAR(45) + CHAR(37), @StringToStrip);
    WHILE @DoubleDashStart > 0
    BEGIN
        -- Search for the first "new line" AFTER the first "double dash"
        -- We can use CHAR(13) and CHAR(10) to find the new line.
        -- Since PATINDEX doesn't have a "start" character, and we need to find
        -- the first new line AFTER the double dash, we will search all characters
        -- after the double dash for the new line.
        SET @DoubleDashLineEnd = PATINDEX('%' + CHAR(10) + '%',
                                        RIGHT(@StringToStrip, LEN(@StringToStrip) - (@DoubleDashStart)))
                                 + @DoubleDashStart;
                                        
        -- "Cut" out the comment, as was done with the block comments.
        SET @StringToStrip =    LEFT(@StringToStrip, @DoubleDashStart - 1) +
                                RIGHT(@StringToStrip, LEN(@StringToStrip) - @DoubleDashLineEnd);
    
        -- Check for the next incidence of a double dash, if there is one.
        SET @DoubleDashStart = PATINDEX(CHAR(37) + CHAR(45) + CHAR(45) + CHAR(37), @StringToStrip);
    END;
    -- Return the uncommented string    
    RETURN @StringToStrip;
END;
GO
IF OBJECT_ID('tempdb.[dbo].[#AllDefinitions]') IS NOT NULL 
DROP TABLE [dbo].[#AllDefinitions]; 
SELECT IDENTITY(INT,1,1) AS [RW],
QUOTENAME(OBJECT_SCHEMA_NAME([object_id])) + '.' + QUOTENAME(OBJECT_NAME([object_id])) AS [QualifiedObjectName],
OBJECT_SCHEMA_NAME([object_id]) AS [SchemaName],
OBJECT_NAME([object_id]) AS [ObjectName],
[object_id],
[definition],
[definition] AS [OriginalDefinition]
INTO [#AllDefinitions]
FROM [sys].[sql_modules];
UPDATE [#AllDefinitions] 
SET [definition] = REPLACE([definition],CHAR(13)+ CHAR(10),CHAR(10));
--standardize the CrLf to plain old LF
--#################################################################################################
-- Remove comments to reduce false positives from commented code
--#################################################################################################

UPDATE [#AllDefinitions]
SET [definition] =  [dbo].[StripComments]([definition]);
--#################################################################################################
 --whitespace: char13/char10/tab/space
UPDATE [#AllDefinitions]
SET [definition] = REPLACE(REPLACE(REPLACE(REPLACE([definition],CHAR(13),''),CHAR(10),''),CHAR(9),''),CHAR(32),'');
DECLARE @SearchTerm VARCHAR(128);
SET @SearchTerm = 'TABLE(';
--A table variable declaration exists?
SELECT @SearchTerm AS [SearchTerm],SUBSTRING([definition],CHARINDEX(@SearchTerm,[definition]) - 15,115),* FROM [#AllDefinitions] WHERE [definition] LIKE '%' + @SearchTerm + '%';
--Top 1
SET @SearchTerm = 'TOP1';
SELECT @SearchTerm AS [SearchTerm],SUBSTRING([definition],CHARINDEX(@SearchTerm,[definition]) - 15,115),* FROM [#AllDefinitions] WHERE [definition] LIKE '%' + @SearchTerm + '%';
--Top using a variable?
SET @SearchTerm = 'TOP(';
SELECT @SearchTerm AS [SearchTerm],SUBSTRING([definition],CHARINDEX(@SearchTerm,[definition]) - 15,115),* FROM [#AllDefinitions] WHERE [definition] LIKE '%' + @SearchTerm + '%';
--NOLOCK hint?
SET @SearchTerm ='NOLOCK';
SELECT @SearchTerm AS [SearchTerm],SUBSTRING([definition],CHARINDEX(@SearchTerm,[definition]) - 15,115),* FROM [#AllDefinitions] WHERE [definition] LIKE '%' + @SearchTerm + '%';
--MERGEJOIN
SET @SearchTerm = 'MERGEJOIN';
SELECT @SearchTerm AS [SearchTerm],SUBSTRING([definition],CHARINDEX(@SearchTerm,[definition]) - 15,115),* FROM [#AllDefinitions] WHERE [definition] LIKE '%' + @SearchTerm + '%';
--%LOOPJOIN
SET @SearchTerm = 'LOOPJOIN';
SELECT @SearchTerm AS [SearchTerm],SUBSTRING([definition],CHARINDEX(@SearchTerm,[definition]) - 15,115),* FROM [#AllDefinitions] WHERE [definition] LIKE '%' + @SearchTerm + '%';
--%LOOPJOIN
SET @SearchTerm = 'HASHJOIN';
SELECT @SearchTerm AS [SearchTerm],SUBSTRING([definition],CHARINDEX(@SearchTerm,[definition]) - 15,115),* FROM [#AllDefinitions] WHERE [definition] LIKE '%' + @SearchTerm + '%';
GO
IF OBJECT_ID('[dbo].[StripComments]') IS NOT NULL 
DROP  FUNCTION [dbo].[StripComments]; 
GO
GO
IF OBJECT_ID('tempdb..#ModuleDefinitions') IS NOT NULL DROP TABLE #ModuleDefinitions;
CREATE TABLE #ModuleDefinitions(
[Observation]   VARCHAR(256)                             NOT NULL,
[DatabaseName]  SYSNAME                                 NOT NULL,
[SchemaName]    SYSNAME                                 NOT NULL,
[ObjectName]     SYSNAME                                 NOT NULL,
[ObjectType]    NVARCHAR(128)                               NULL,
[Definition]    NVARCHAR(max)                               NULL)

EXECUTE sp_MSforeachdb '
INSERT INTO #ModuleDefinitions([Observation],[DatabaseName],[SchemaName],[ObjectName],[ObjectType],[Definition])
SELECT ''Nested CTE Review'' AS Observation,
  ''?'' AS DatabaseName,
  schmz.name As SchemaName,
  tabz.name AS [ObjectName],
  tabz.type_desc AS ObjectType,
  def.[definition]
FROM [?].sys.objects tabz
  INNER JOIN [?].sys.schemas schmz ON tabz.schema_id = schmz.schema_id
  INNER JOIN [?].sys.sql_modules def  ON tabz.object_id = def.object_id
'
--delete from #ModuleDefinitions WHERE DatabaseName IN('master','msdb')

-- Update #ModuleDefinitions to strip multi-line comments
WHILE EXISTS(SELECT * FROM #ModuleDefinitions WHERE  CHARINDEX('/*', Definition) > 0 AND CHARINDEX('*/', Definition) > CHARINDEX('/*', Definition))
	BEGIN
		PRINT 'Cleaning multi line comments'
		UPDATE #ModuleDefinitions
		SET Definition = CASE 
			WHEN CHARINDEX('/*', Definition) > 0 AND CHARINDEX('*/', Definition) > CHARINDEX('/*', Definition)
			THEN STUFF(
				Definition,
				CHARINDEX('/*', Definition),
				CHARINDEX('*/', Definition) - CHARINDEX('/*', Definition) + 2,
				''
			)
			ELSE Definition
		END
		WHERE CHARINDEX('/*', Definition) > 0 
	END
-- Update #ModuleDefinitions to strip single-line comments; this is a little slow, as it might loop procedures hundreds of times, depending on comments
WHILE EXISTS(SELECT * FROM #ModuleDefinitions WHERE  CHARINDEX('--', Definition) > 0 )
	BEGIN
		UPDATE #ModuleDefinitions
		SET Definition = CASE 
			WHEN CHARINDEX('--', Definition) > 0 
			THEN STUFF(
				Definition,
				CHARINDEX('--', Definition),
				CHARINDEX(CHAR(10), Definition + CHAR(10), CHARINDEX('--', Definition)) - CHARINDEX('--', Definition),
				''
			)
			ELSE Definition
		END
		WHERE  CHARINDEX('--', Definition) > 0
	END
UPDATE #ModuleDefinitions SET definition = REPLACE(definition, CHAR(13) + CHAR(10), ' '); -- Remove line breaks
UPDATE #ModuleDefinitions SET definition = REPLACE(definition, CHAR(9), ' '); -- Remove tabs
UPDATE #ModuleDefinitions SET definition = REPLACE(definition, CHAR(13) , ' '); -- Remove line breaks
UPDATE #ModuleDefinitions SET definition = REPLACE(definition, CHAR(10), ' '); -- Remove line breaks
UPDATE #ModuleDefinitions SET definition = REPLACE(definition, ' ', ''); -- Remove all spaces
-- Query to find modules with nested CTEs using CROSS APPLY
DECLARE @SearchTerm TABLE (SearchTerm VARCHAR(128), Observation VARCHAR(256));
INSERT INTO @SearchTerm VALUES	('TABLE(','@Table Variables have no statistics; when used improperly with larger rowcounts, they are an easy performance target to switch to #temptables'),
								('TOP1','TOP 1 is an sometimes an indicator of a refactor possibility of either a missing join with group by, or an exists operator'),
								('TOP(','TOP(@num still needs to be reviewed for the same TOP 1 opportunity to refactor missing join with group by, or an exists operator'),
								('NOLOCK','the NOLOCK hint, if deemed appropriate, could be replaced with transaction isolation level read uncommitted. this would make it easier in the future if the code needs to be changed to snapshot or some other isolation level '),
								('MERGEJOIN','Overriding the ability for the SQL engine to decide the proper join type should be rarely done, as outside factors like new indexes and proper statistics makes this 99.9% unnesssary)'),
								('LOOPJOIN','Overriding the ability for the SQL engine to decide the proper join type should be rarely done, as outside factors like new indexes and proper statistics makes this 99.9% unnesssary)'),
								('HASHJOIN','Overriding the ability for the SQL engine to decide the proper join type should be rarely done, as outside factors like new indexes and proper statistics makes this 99.9% unnesssary)')
SELECT md.[DatabaseName],
st.SearchTerm AS [SearchTerm],
st.[Observation],
SUBSTRING(md.[definition],
CHARINDEX(st.SearchTerm,md.[definition]) - 15,115)AS Overview,
md.[SchemaName],
md.[ObjectName],
md.[ObjectType],
md.[Definition]
FROM #ModuleDefinitions md
INNER JOIN @SearchTerm st ON md.[definition] LIKE '%' + st.SearchTerm + '%'
WHERE md.[DatabaseName] NOT IN('master','msdb')
GO
